from functools import partial

from pydantic import ValidationError
from pytest import raises

from stjames import Atom, Molecule


def test_molecule_pbc() -> None:
    # this is a funny joke, because it makes manganese
    glomar_explorer = partial(Molecule, charge=0, multiplicity=2, atoms=[Atom(atomic_number=25, position=[0, 0, 0])])

    mol_nopbc = glomar_explorer()
    assert mol_nopbc.cell is None

    valid_input = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))
    mol = glomar_explorer(cell={"lattice_vectors": valid_input})
    assert mol.cell is not None
    assert mol.cell.lattice_vectors == valid_input

    mol2 = glomar_explorer(cell={"lattice_vectors": valid_input, "is_periodic": (True, False, True)})
    assert mol2.cell is not None
    assert mol2.cell.lattice_vectors == valid_input
    assert mol2.cell.is_periodic == (True, False, True)

    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": valid_input, "is_periodic": (False, False, False)})

    invalid_input = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input})

    invalid_input2 = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, "invalid"))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input2})

    invalid_input3 = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0), (1.0, 1.0, 1.0))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input3})



"""ADME-Tox property prediction workflow."""

import warnings
from typing import Self

from pydantic import model_validator

from ..molecule import Molecule
from .workflow import MoleculeWorkflow, SMILESWorkflow


class ADMETWorkflow(SMILESWorkflow, MoleculeWorkflow):
    """
    A workflow for predicting ADME-Tox properties.

    Inherited:
    :param initial_smiles: SMILES string of molecule (mutually exclusive with initial_molecule)
    :param initial_molecule: Molecule of interest (deprecated)
    :param mode: Mode for workflow (currently unused)

    New:
    :param properties: predicted properties
    """

    initial_smiles: str = ""
    initial_molecule: Molecule | None = None  # type: ignore [assignment]  # Deprecated
    properties: dict[str, float | int] | None = None

    @model_validator(mode="after")
    def validate_mol_input(self) -> Self:
        """Ensure that only one of initial_molecule or initial_smiles is set."""

        if not (bool(self.initial_smiles) ^ bool(self.initial_molecule)):
            raise ValueError("Can only set one of initial_molecule should and initial_smiles")

        if self.initial_molecule is not None:
            warnings.warn(DeprecationWarning("initial_molecule is deprecated. Use initial_smiles instead."))

        return self



from enum import Enum
from typing import Annotated, Any, Callable, Hashable, Optional, TypeVar

import numpy as np
import pydantic

_T = TypeVar("_T")


def round_float(round_to: int) -> Callable[[float], float]:
    """Return a function that rounds a float to a given number of decimal places."""

    def inner_round(v: float) -> float:
        return round(v, round_to)

    return inner_round


def round_optional_float(round_to: int) -> Callable[[Optional[float]], Optional[float]]:
    """Create a validator that rounds an optional float to a given number of decimal places."""

    def rounder(value: Optional[float]) -> Optional[float]:
        if value is None:
            return None
        return round(value, round_to)

    return rounder


class Base(pydantic.BaseModel):
    @pydantic.field_validator("*", mode="before")
    @classmethod
    def coerce_numpy(cls, val: _T) -> _T | list[Any]:
        if isinstance(val, np.ndarray):
            return val.tolist()  # type: ignore [no-any-return, unused-ignore, return-value]

        return val


class LowercaseStrEnum(str, Enum):
    """Enum where hyphens, underscores, and case are ignored."""

    @classmethod
    def _missing_(cls, value: object) -> str | None:
        for member in cls:
            if isinstance(value, str):
                if member.lower().replace("-", "").replace("_", "") == value.lower().replace("-", "").replace("_", ""):
                    return member
        return None


# cf. https://github.com/pydantic/pydantic-core/pull/820#issuecomment-1670475909
_H = TypeVar("_H", bound=Hashable)


def _validate_unique_list(v: list[_H]) -> list[_H]:
    if len(v) != len(set(v)):
        raise ValueError("this list must be unique, and isn't!")
    return v


UniqueList = Annotated[list[_H], pydantic.AfterValidator(_validate_unique_list)]



"""Hydrogen-bond-basicity workflow."""

from ..base import Base
from ..types import UUID
from .workflow import MoleculeWorkflow


class HydrogenBondAcceptorSite(Base):
    """
    A hydrogen bond acceptor site.

    :param atom_idx: index of the atom
    :param pkbhx: Hydrogen bond basicity
    :param position: position of the atom
    :param name: name of the atom
    """

    atom_idx: int  # zero-indexed
    pkbhx: float
    position: tuple[float, float, float]
    name: str | None = None


class HydrogenBondBasicityWorkflow(MoleculeWorkflow):
    """
    Workflow for calculating hydrogen bond basicity.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param do_csearch: whether to perform a conformational search
    :param do_optimization: whether to perform an optimization

    Results:
    :param optimization: UUID of optimization
    :param hba_sites: hydrogen-bond-acceptor sites
    """

    do_csearch: bool = True
    do_optimization: bool = True

    optimization: UUID | None = None
    hba_sites: list[HydrogenBondAcceptorSite] = []  # noqa: RUF012



"""Tautomer prediction workflow."""

from typing import Annotated, Optional

from pydantic import AfterValidator

from ..base import Base, round_float, round_optional_float
from ..mode import Mode
from .workflow import DBCalculation, MoleculeWorkflow


class Tautomer(Base):
    """
    A tautomer.

    :param energy: energy of the tautomer
    :param weight: statistical weight of the tautomer
    :param predicted_relative_energy: relative energy of the tautomer
    :param structures: UUIDs of the structures
    """

    energy: Annotated[float, AfterValidator(round_float(6))]
    weight: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None
    predicted_relative_energy: Annotated[Optional[float], AfterValidator(round_optional_float(6))] = None

    # UUIDs, optionally
    structures: list[DBCalculation] = []


class TautomerWorkflow(MoleculeWorkflow):
    """
    A workflow to calculate tautomers.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow

    Results:
    :param tautomers: resulting Tautomers
    """

    mode: Mode = Mode.CAREFUL
    tautomers: list[Tautomer] = []



"""Fukui index workflow."""

from typing import Annotated

from pydantic import AfterValidator

from ..base import round_optional_float
from ..types import UUID, FloatPerAtom, round_optional_float_per_atom
from .workflow import MoleculeWorkflow


class FukuiIndexWorkflow(MoleculeWorkflow):
    """
    Workflow for calculating Fukui indices.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    Results:
    :param optimization: UUID of optimization
    :param global_electrophilicity_index: global electrophilicity index
    :param fukui_positive: Fukui index for positive charges
    :param fukui_negative: Fukui index for negative charges
    :param fukui_zero: Fukui index for zero charges
    """

    # UUID of optimization
    optimization: UUID | None = None

    global_electrophilicity_index: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    fukui_positive: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    fukui_negative: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
    fukui_zero: Annotated[FloatPerAtom | None, AfterValidator(round_optional_float_per_atom(6))] = None
